package index

import (
	"squirreldb/types"
	"sync"
	"time"

	"github.com/pilosa/pilosa/v2/roaring"
	"github.com/prometheus/prometheus/model/labels"
)

// labelsLookupCache provide cache for metricID to labels.
// It's eviction policy is random (Golang map order).
type labelsLookupCache struct {
	cache map[types.MetricID]labelsEntry
	l     sync.Mutex
}

// postingsCache provide a cache for postings queries by label=value.
// It is only used for query and not metric creation, mostly because creation only
// occure once thus a cache is not useful.
// Invalidation occur from TTL and on metric creation/deletion: a message is sent to all
// SquirrelDB and posting that match the created metric are invalidated.
type postingsCache struct {
	// cache map shard => label name => label value => postingEntry
	cache map[postingsCacheKey]postingEntry
	l     sync.Mutex
}

type postingsCacheKey struct {
	Name  string
	Value string
	Shard int32
}

type postingEntry struct {
	expire time.Time
	value  *roaring.Bitmap
	count  uint64
}

type labelsEntry struct {
	cassandraExpire time.Time
	value           labels.Labels
}

const (
	labelCacheMaxSize    = 10000
	postingsCacheMaxSize = 10000
	postingsCacheTTL     = 15 * time.Minute
)

// Get return the non-expired cache entry or an nil list.
func (c *labelsLookupCache) Get(now time.Time, id types.MetricID) labels.Labels {
	c.l.Lock()
	defer c.l.Unlock()

	return c.get(now, id)
}

// MGet return the non-expired cache entry or an nil list for multiple IDs.
func (c *labelsLookupCache) MGet(now time.Time, ids []types.MetricID) []labels.Labels {
	c.l.Lock()
	defer c.l.Unlock()

	result := make([]labels.Labels, len(ids))
	for i, id := range ids {
		result[i] = c.get(now, id)
	}

	return result
}

// Set add entry to the cache. Return the current cache size.
func (c *labelsLookupCache) Set(
	now time.Time,
	id types.MetricID,
	value labels.Labels,
	cassandraExpiration time.Time,
) int {
	c.l.Lock()
	defer c.l.Unlock()

	if len(c.cache) > labelCacheMaxSize {
		// First drop expired entry
		for k, v := range c.cache {
			if v.cassandraExpire.Before(now) {
				delete(c.cache, k)
			}
		}

		// we want to evict at least 50%, to avoid doing set/evict one/set/evict one/...
		if len(c.cache) > labelCacheMaxSize/2 {
			toDelete := len(c.cache) - labelCacheMaxSize/2

			for k := range c.cache {
				delete(c.cache, k)

				toDelete--

				if toDelete <= 0 {
					break
				}
			}
		}
	}

	c.cache[id] = labelsEntry{
		value:           value,
		cassandraExpire: cassandraExpiration,
	}

	return len(c.cache)
}

// Drop delete entries from cache. Return the cache size.
func (c *labelsLookupCache) Drop(ids []uint64) int {
	c.l.Lock()
	defer c.l.Unlock()

	for _, id := range ids {
		delete(c.cache, types.MetricID(id))
	}

	return len(c.cache)
}

func (c *labelsLookupCache) get(now time.Time, id types.MetricID) labels.Labels {
	entry := c.cache[id]
	if entry.cassandraExpire.IsZero() {
		return nil
	}

	if entry.cassandraExpire.Before(now) {
		delete(c.cache, id)

		return nil
	}

	return entry.value
}

// Get return the non-expired cache entry or nil.
func (c *postingsCache) Get(shard int32, name string, value string) *roaring.Bitmap {
	return c.get(time.Now(), shard, name, value)
}

// Invalidate drop entry that are impacted by given labels. Return the cache size.
func (c *postingsCache) Invalidate(entries []postingsCacheKey) int {
	c.l.Lock()
	defer c.l.Unlock()

	for _, k := range entries {
		delete(c.cache, k)
	}

	return len(c.cache)
}

// Set add an entry. Return the cache size.
func (c *postingsCache) Set(shard int32, name string, value string, bitmap *roaring.Bitmap) int {
	now := time.Now()

	return c.set(now, shard, name, value, bitmap)
}

func (c *postingsCache) set(now time.Time, shard int32, name string, value string, bitmap *roaring.Bitmap) int {
	c.l.Lock()
	defer c.l.Unlock()

	if len(c.cache) > postingsCacheMaxSize {
		// First drop expired entry
		for k, v := range c.cache {
			if v.expire.Before(now) {
				delete(c.cache, k)
			}
		}

		// we want to evict at least 50%, to avoid doing set/evict one/set/evict one/...
		if len(c.cache) > postingsCacheMaxSize/2 {
			toDelete := len(c.cache) - postingsCacheMaxSize/2

			for k := range c.cache {
				delete(c.cache, k)

				toDelete--

				if toDelete <= 0 {
					break
				}
			}
		}
	}

	key := postingsCacheKey{
		Shard: shard,
		Name:  name,
		Value: value,
	}

	c.cache[key] = postingEntry{
		expire: now.Add(postingsCacheTTL),
		value:  bitmap,
		count:  bitmap.Count(),
	}

	return len(c.cache)
}

func (c *postingsCache) get(now time.Time, shard int32, name string, value string) *roaring.Bitmap {
	c.l.Lock()
	defer c.l.Unlock()

	key := postingsCacheKey{
		Shard: shard,
		Name:  name,
		Value: value,
	}

	entry := c.cache[key]

	if entry.expire.IsZero() {
		return nil
	}

	if entry.value.Count() != entry.count {
		panic("fuck...")
	}

	if entry.expire.Before(now) {
		delete(c.cache, key)

		return nil
	}

	return entry.value
}
